
#################################
# YUIMAチュートリアルオープニング
#################################

#################################
# 1 YUIMA p.1 （レジュメ）
# YUIMA（ユイマ）とは
# • 確率過程に対する統計解析および
#   シミュレーションのためのパッケージ
# • R 上に実装されている
# • 確率微分方程式や点過程をオブジェクト化する
# • 最先端の確率統計学の理論に基づく結果を
#   容易に利用できる


#################################
# 2 一つの例 p.1 （レジュメ）
# YUIMA を使った統計解析の例を一つあげる。

# コード等の詳細はチュートリアルで説明するので、
# 今はどのようなことができるか眺めるだけで良い。

# ここでのタスクは
# 1. YUIMA をインストール
# 2. 株価データをダウンロード
# 3. 確率モデルをデータに当てはめる
# 4. 当てはめたモデルを用いて将来の値動きを
#     シミュレーションで予測する
# である。

#################################
# 3 YUIMAの準備 p.1 
# Rをインストールする。
# RStudioをインストールする。
# YUIMA パッケージをインストールする。
# install.packages("yuima") "#"はコメントアウト。外すとそこが実行される。
# require(yuima) # パッケージ yuima をロードする。


#################################
# 4 データオブジェクトの作成 p.2

# 株価データを取得するために
# quantmod(Quantitative Financial Modelling Framework) 
# パッケージをインストールし、読み込む:
# install.packages("quantmod")
# quantmod(Quantitative Financial Modelling Framework)パッケージをロード
require(quantmod) 

# IBM のデータをダウンロード:
getSymbols("IBM", from = "2016-01-01", to = "2017-12-31")

# データ"IBM"の構造を見てみる：
# str(IBM) # Yahoo Financeからダウンロードされたデータが
#          xts (time series) オブジェクトIBMに変換されている。

#  データの始めと終わりを見てみる:
head(IBM)
tail(IBM)
# IBM2years <- IBM データのバックアップ


#################################
# YUIMAを使う！

# YUIMAオブジェクトの構造 （レジュメ p.8の図参照）
#    model 
#    data
#    sampling 
#    .....
#    .....
# YUIMAオブジェクトを"場"として、解析を行う。

#################################
# ↓ setDate, setYuima で
# データオブジェクト、
# YUIMAオブジェクトをつくる p.2：
x <- setYuima(data=setData(IBM$IBM.Close)) 
str(x@data) # xにIBM.Closeデータが確認できる
# プロットしてみる。
plot(x, main="data with the original time stamps")


#################################
# 5 モデルのフィット p.3

# YUIMAでデータにBlack-Scholesモデル
# をあてはめる。

#################################
# まず、時間スケールの標準化 
# ↓ 時間の単位として１年を1単位時間とする。
# Δ = 1/252 = 0.00397、252は１年のワーキングデーの数
# YUIMAオブジェクトを作る：
y <- setYuima(data=setData(IBM$IBM.Close, delta=1/252))
# str(y@data) #１ステップの時間が0.00397となった。
# ↓ 両者をプロットして比較
par(mfrow=c(1,2)) # Plotsペインを広げておくこと。
plot(x, main="data with the original time stamps")
plot(y, main="time stamps of data rescaled")
par(mfrow=c(1,1))
plot(y, main="time stamps of data rescaled")
# ↓ x,yのサマリー。deltaの値が異なることに注意。
# x
# y


#################################
# Black-Scholes モデル 
#   dX_t = μX_tdt + σX_tdWt
# のモデルオブジェクトを作る p.4：
IBM.mod <- setModel(
  drift=c("mu*x"),diffusion=matrix(c("sigma*x"),1,1))
# str(IBM.mod)
IBM.mod.data <- setYuima(model=IBM.mod, data=y@data)
str(IBM.mod.data) 
# ↑ データスロットとモデルスロットが
# 埋まっていることを確認。
# plot(IBM.mod.data) # プロットもできる。

#################################
# モデルのデータへのフィット p.4
IBM.fit <- qmle(IBM.mod.data,
                lower=list(mu=0.01, sigma=0.01),
                start=list(mu=1, sigma=1),
                upper=list(mu=10, sigma=10), 
                method="L-BFGS-B")
# str(IBM.fit)
IBM.fit@coef # パラメータの推定値


#################################
# 6 シミュレーションによる
#   IBMのパスの将来予測 p.5
#
# 先ほどダウンロードしたデータは
# 2016-01-04 から2017-12-29 であった。

# ２０１８年はじめから約３ヶ月先までの
# 株価のパスをシミュレーションで予測する。

# 後の比較に用いるデータを
# ダウンロードする。
getSymbols("IBM", from = "2018-01-02", to = "2018-04-05")
IBM.first <- IBM$IBM.Close["2018-01-02"]
# 年の初めは2018-01-02である。
# IBM3months <- IBM # バックアップ


# 予測（シミュレーション）のために
# 0.25 年先までのモデルを設定する：
IBM.mod.pred <- setYuima(model=IBM.mod, 
                 sampling = setSampling(Terminal=0.25, n=250))
# str(IBM.mod.pred)
# 推定したモデルを使って
# パスを生成してプロットする。
# IBM.fit@coef # BSモデルのパラメータ推定値
set.seed(127)
IBM.pred <- simulate(IBM.mod.pred, 
                     true.parameter = list(sigma=IBM.fit@coef[1],
                                           mu=IBM.fit@coef[2]), 
                     xinit = IBM.first)
plot(IBM.pred, ylim=c(100,220), xlim=c(0,0.27), col="skyblue", 
     ylab=c("IBM.Close"),xlab=c("t"))


# １本のパスでは傾向がわからないので、
#１００本のパスをモデルから生成して
# 重ねてプロットする。
# 最後に実際のパスを重ねてみる:

## シミュレーターの作成
mypred <- function(param){
  result <- simulate(IBM.mod.pred,
              true.parameter = list(sigma=param[1], mu=param[2]),
              xinit = IBM.first)
  return(get.zoo.data(result)[[1]])
}

## 100本のパスをシミュレートして
## 結果を保存:
path <- replicate(100, mypred(IBM.fit@coef))

## シミュレートしたパスをプロット:
time.idx <- IBM.mod.pred@sampling@grid[[1]] # タイムインデックス
matplot(time.idx, path, type="l", lty=1, col="skyblue",
        xlab="t", ylab="IBM.Close", axes=FALSE,
        main="IBM 2018.01.02-2018.04.04")

## 実際のパスを上書き
IBM2018.y <- setYuima(data=setData(IBM$IBM.Close, delta=1/252))
IBM2018.y <- get.zoo.data(IBM2018.y)[[1]]
lines(time(IBM2018.y),IBM2018.y, col="red",lwd=2)
## 軸の設定
idx <- seq(1, length(time(IBM)), by = 9)
axis(1, at = time(IBM2018.y)[idx], labels = time(IBM)[idx])
axis(2)

#################################
# 「一般的な注意」を読んでご理解ください。


#################################
# 時間割
# 1日目 (12月1日, 大講義室)：
# 10:30-10:50 チュートリアル概観 （吉田）
# 11:00-12:00 ブラウン運動と確率解析入門（小池）
# 13:30-15:00 確率微分方程式入門 （小池）
# 15:30-17:00 モデルパラメータの推定 （上原）

# 2日目 (12月2日, 大講義室)：
# 10:30-12:00 モデル選択 （江口）
# 13:30-14:30 レヴィ過程 （増田）
# 14:50-15:50 デフォルト・破産確率の計算（清水）
# 15:50-16:00 終わりの挨拶、解散
# 16:00-17:00 フリーディスカッション
#################################
